@@ -11,7 +11,8 @@ from .pdu import PCI, PDUData
 from .primitivedata import Boolean, CharacterString, Enumerated, Integer, \
     ObjectIdentifier, ObjectType, OctetString, Real, TagList, Unsigned, \
     expand_enumerations
-from .constructeddata import Any, Choice, Element, Sequence, SequenceOf
+from .constructeddata import Any, Choice, Element, Sequence, SequenceOf, \
+    SequenceOfAny
 from .basetypes import ChannelValue, DateTime, DeviceAddress, ErrorType, \
     EventState, EventTransitionBits, EventType, LifeSafetyOperation, \
     NotificationParameters, NotifyType, ObjectPropertyReference, \
@@ -56,38 +57,51 @@ def register_error_type(klass):
 #   encode_max_segments_accepted/decode_max_segments_accepted
 #
 
+_max_segments_accepted_encoding = [
+    None, 2, 4, 8, 16, 32, 64, None,
+    ]
+
 def encode_max_segments_accepted(arg):
     """Encode the maximum number of segments the device will accept, Section
-    20.1.2.4"""
-    w = 0
-    while (arg and not arg & 1):
-        w += 1
-        arg = (arg >> 1)
-    return w
+    20.1.2.4, and if the device says it can only accept one segment it shouldn't
+    say that it supports segmentation!"""
+    # unspecified
+    if not arg:
+        return 0
+
+    if arg > 64:
+        return 7
+
+    # the largest number not greater than the arg
+    for i in range(6, 0, -1):
+        if _max_segments_accepted_encoding[i] <= arg:
+            return i
+
+    raise ValueError("invalid max max segments accepted: {0}".format(arg))
 
 def decode_max_segments_accepted(arg):
     """Decode the maximum number of segments the device will accept, Section
     20.1.2.4"""
-    return arg and (1 << arg) or None
+    return _max_segments_accepted_encoding[arg]
 
 #
 #   encode_max_apdu_length_accepted/decode_max_apdu_length_accepted
 #
 
-_max_apdu_response_encoding = [50, 128, 206, 480, 1024, 1476, None, None,
+_max_apdu_length_encoding = [50, 128, 206, 480, 1024, 1476, None, None,
     None, None, None, None, None, None, None, None]
 
 def encode_max_apdu_length_accepted(arg):
     """Return the encoding of the highest encodable value less than the
     value of the arg."""
     for i in range(5, -1, -1):
-        if (arg >= _max_apdu_response_encoding[i]):
+        if (arg >= _max_apdu_length_encoding[i]):
             return i
 
     raise ValueError("invalid max APDU length accepted: {0}".format(arg))
 
 def decode_max_apdu_length_accepted(arg):
-    v = _max_apdu_response_encoding[arg]
+    v = _max_apdu_length_encoding[arg]
     if not v:
         raise ValueError("invalid max APDU length accepted: {0}".format(arg))
 
@@ -174,7 +188,7 @@ class APCI(PCI, DebugContents):
             if self.apduSA:
                 buff += 0x02
             pdu.put(buff)
-            pdu.put((encode_max_segments_accepted(self.apduMaxSegs) << 4) + encode_max_apdu_length_accepted(self.apduMaxResp))
+            pdu.put((self.apduMaxSegs << 4) + self.apduMaxResp)
             pdu.put(self.apduInvokeID)
             if self.apduSeg:
                 pdu.put(self.apduSeq)
@@ -255,8 +269,8 @@ class APCI(PCI, DebugContents):
             self.apduMor = ((buff & 0x04) != 0)
             self.apduSA  = ((buff & 0x02) != 0)
             buff = pdu.get()
-            self.apduMaxSegs = decode_max_segments_accepted( (buff >> 4) & 0x07 )
-            self.apduMaxResp = decode_max_apdu_length_accepted( buff & 0x0F )
+            self.apduMaxSegs = (buff >> 4) & 0x07
+            self.apduMaxResp = buff & 0x0F
             self.apduInvokeID = pdu.get()
             if self.apduSeg:
                 self.apduSeq = pdu.get()
@@ -349,6 +363,7 @@ class APCI(PCI, DebugContents):
 #   APDU
 #
 
+@bacpypes_debugging
 class APDU(APCI, PDUData):
 
     def __init__(self, *args, **kwargs):
@@ -356,12 +371,13 @@ class APDU(APCI, PDUData):
         super(APDU, self).__init__(*args, **kwargs)
 
     def encode(self, pdu):
-        if _debug: APCI._debug("encode %s", str(pdu))
+        if _debug: APDU._debug("encode %s", str(pdu))
         APCI.encode(self, pdu)
         pdu.put_data(self.pduData)
 
     def decode(self, pdu):
-        if _debug: APCI._debug("decode %s", str(pdu))
+        if _debug: APDU._debug("decode %s", str(pdu))
+
         APCI.decode(self, pdu)
         self.pduData = pdu.get_data(len(pdu.pduData))
 
@@ -393,17 +409,24 @@ class APDU(APCI, PDUData):
 #   between PDU's.  Otherwise the APCI content would be decoded twice.
 #
 
+@bacpypes_debugging
 class _APDU(APDU):
 
     def encode(self, pdu):
+        if _debug: _APDU._debug("encode %r", pdu)
+
         APCI.update(pdu, self)
         pdu.put_data(self.pduData)
 
     def decode(self, pdu):
+        if _debug: _APDU._debug("decode %r", pdu)
+
         APCI.update(self, pdu)
         self.pduData = pdu.get_data(len(pdu.pduData))
 
     def set_context(self, context):
+        if _debug: _APDU._debug("set_context %r", context)
+
         self.pduUserData = context.pduUserData
         self.pduDestination = context.pduSource
         self.pduExpectingReply = 0
@@ -898,7 +921,7 @@ class RangeByPosition(Sequence):
 
 class RangeBySequenceNumber(Sequence):
     sequenceElements = \
-        [ Element('referenceIndex', Unsigned)
+        [ Element('referenceSequenceNumber', Unsigned)
         , Element('count', Integer)
         ]
 
@@ -934,7 +957,7 @@ class ReadRangeACK(ComplexAckSequence):
         , Element('propertyArrayIndex', Unsigned, 2, True)
         , Element('resultFlags', ResultFlags, 3)
         , Element('itemCount', Unsigned, 4)
-        , Element('itemData', SequenceOf(Any), 5)
+        , Element('itemData', SequenceOfAny, 5)
         , Element('firstSequenceNumber', Unsigned, 6, True)
         ]
 
@@ -1292,7 +1315,7 @@ class LifeSafetyOperationRequest(ConfirmedRequestSequence):
         [ Element('requestingProcessIdentifier', Unsigned, 0)
         , Element('requestingSource', CharacterString, 1)
         , Element('request', LifeSafetyOperation, 2)
-        , Element('objectIdentifier', ObjectIdentifier, 3)
+        , Element('objectIdentifier', ObjectIdentifier, 3, True)
         ]
 
 register_confirmed_request_type(LifeSafetyOperationRequest)
@@ -1476,7 +1499,7 @@ class RemoveListElementRequest(ConfirmedRequestSequence):
     sequenceElements = \
         [ Element('objectIdentifier', ObjectIdentifier, 0)
         , Element('propertyIdentifier', PropertyIdentifier, 1)
-        , Element('propertyArrayIndex', Unsigned, 2)
+        , Element('propertyArrayIndex', Unsigned, 2, True)
         , Element('listOfElements', Any, 3)
         ]
 
