@@ -26,7 +26,8 @@ from .errors import ExecutionError, UnrecognizedService, AbortException, RejectE
 
 # for computing protocol services supported
 from .apdu import confirmed_request_types, unconfirmed_request_types, \
-    ConfirmedServiceChoice, UnconfirmedServiceChoice
+    ConfirmedServiceChoice, UnconfirmedServiceChoice, \
+    IAmRequest
 from .basetypes import ServicesSupported
 
 # basic services
@@ -54,16 +55,16 @@ class DeviceInfo(DebugContents):
         'maxSegmentsAccepted',
         )
 
-    def __init__(self):
+    def __init__(self, device_identifier, address):
         # this information is from an IAmRequest
-        self.deviceIdentifier = None                    # device identifier
-        self.address = None                             # LocalStation or RemoteStation
+        self.deviceIdentifier = device_identifier
+        self.address = address
+
         self.maxApduLengthAccepted = 1024               # maximum APDU device will accept
         self.segmentationSupported = 'noSegmentation'   # normally no segmentation
+        self.maxSegmentsAccepted = None                 # None iff no segmentation
         self.vendorID = None                            # vendor identifier
-
-        self.maxNpduLength = 1497           # maximum we can send in transit
-        self.maxSegmentsAccepted = None     # value for proposed/actual window size
+        self.maxNpduLength = None           # maximum we can send in transit (see 19.4)
 
 #
 #   DeviceInfoCache
@@ -72,56 +73,105 @@ class DeviceInfo(DebugContents):
 @bacpypes_debugging
 class DeviceInfoCache:
 
-    def __init__(self):
+    def __init__(self, device_info_class=DeviceInfo):
         if _debug: DeviceInfoCache._debug("__init__")
 
+        # a little error checking
+        if not issubclass(device_info_class, DeviceInfo):
+            raise ValueError("not a DeviceInfo subclass: %r" % (device_info_class,))
+
         # empty cache
         self.cache = {}
 
+        # class for new records
+        self.device_info_class = device_info_class
+
     def has_device_info(self, key):
         """Return true iff cache has information about the device."""
         if _debug: DeviceInfoCache._debug("has_device_info %r", key)
 
         return key in self.cache
 
-    def add_device_info(self, apdu):
+    def iam_device_info(self, apdu):
         """Create a device information record based on the contents of an
         IAmRequest and put it in the cache."""
-        if _debug: DeviceInfoCache._debug("add_device_info %r", apdu)
+        if _debug: DeviceInfoCache._debug("iam_device_info %r", apdu)
 
-        # get the existing cache record by identifier
-        info = self.get_device_info(apdu.iAmDeviceIdentifier[1])
-        if _debug: DeviceInfoCache._debug("    - info: %r", info)
+        # make sure the apdu is an I-Am
+        if not isinstance(apdu, IAmRequest):
+            raise ValueError("not an IAmRequest: %r" % (apdu,))
 
-        # update existing record
-        if info:
-            if (info.address == apdu.pduSource):
-                return
+        # get the device instance
+        device_instance = apdu.iAmDeviceIdentifier[1]
 
-            info.address = apdu.pduSource
-        else:
-            # get the existing record by address (creates a new record)
-            info = self.get_device_info(apdu.pduSource)
-            if _debug: DeviceInfoCache._debug("    - info: %r", info)
+        # get the existing cache record if it exists
+        device_info = self.cache.get(device_instance, None)
 
-            info.deviceIdentifier = apdu.iAmDeviceIdentifier[1]
+        # maybe there is a record for this address
+        if not device_info:
+            device_info = self.cache.get(apdu.pduSource, None)
 
-        # update the rest of the values
-        info.maxApduLengthAccepted = apdu.maxAPDULengthAccepted
-        info.segmentationSupported = apdu.segmentationSupported
-        info.vendorID = apdu.vendorID
+        # make a new one using the class provided
+        if not device_info:
+            device_info = self.device_info_class(device_instance, apdu.pduSource)
 
-        # say this is an updated record
-        self.update_device_info(info)
+        # jam in the correct values
+        device_info.deviceIdentifier = device_instance
+        device_info.address = apdu.pduSource
+        device_info.maxApduLengthAccepted = apdu.maxAPDULengthAccepted
+        device_info.segmentationSupported = apdu.segmentationSupported
+        device_info.vendorID = apdu.vendorID
+
+        # tell the cache this is an updated record
+        self.update_device_info(device_info)
 
     def get_device_info(self, key):
-        """Return the known information about the device.  If the key is the
-        address of an unknown device, build a generic device information record
-        add put it in the cache."""
         if _debug: DeviceInfoCache._debug("get_device_info %r", key)
 
+        # get the info if it's there
+        device_info = self.cache.get(key, None)
+        if _debug: DeviceInfoCache._debug("    - device_info: %r", device_info)
+
+        return device_info
+
+    def update_device_info(self, device_info):
+        """The application has updated one or more fields in the device
+        information record and the cache needs to be updated to reflect the
+        changes.  If this is a cached version of a persistent record then this
+        is the opportunity to update the database."""
+        if _debug: DeviceInfoCache._debug("update_device_info %r", device_info)
+
+        # give this a reference count if it doesn't have one
+        if not hasattr(device_info, '_ref_count'):
+            device_info._ref_count = 0
+
+        # get the current keys
+        cache_id, cache_address = getattr(device_info, '_cache_keys', (None, None))
+
+        if (cache_id is not None) and (device_info.deviceIdentifier != cache_id):
+            if _debug: DeviceInfoCache._debug("    - device identifier updated")
+
+            # remove the old reference, add the new one
+            del self.cache[cache_id]
+            self.cache[device_info.deviceIdentifier] = device_info
+
+        if (cache_address is not None) and (device_info.address != cache_address):
+            if _debug: DeviceInfoCache._debug("    - device address updated")
+
+            # remove the old reference, add the new one
+            del self.cache[cache_address]
+            self.cache[device_info.address] = device_info
+
+        # update the keys
+        device_info._cache_keys = (device_info.deviceIdentifier, device_info.address)
+
+    def acquire(self, key):
+        """Return the known information about the device and mark the record
+        as being used by a segmenation state machine."""
+        if _debug: DeviceInfoCache._debug("acquire %r", key)
+
         if isinstance(key, int):
-            current_info = self.cache.get(key, None)
+            device_info = self.cache.get(key, None)
 
         elif not isinstance(key, Address):
             raise TypeError("key must be integer or an address")
@@ -130,69 +180,27 @@ class DeviceInfoCache:
             raise TypeError("address must be a local or remote station")
 
         else:
-            current_info = self.cache.get(key, None)
-            if not current_info:
-                current_info = DeviceInfo()
-                current_info.address = key
-                current_info._cache_keys = (None, key)
-                current_info._ref_count = 1
+            device_info = self.cache.get(key, None)
 
-                self.cache[key] = current_info
-            else:
-                if _debug: DeviceInfoCache._debug("    - reference bump")
-                current_info._ref_count += 1
+        if device_info:
+            if _debug: DeviceInfoCache._debug("    - reference bump")
+            device_info._ref_count += 1
 
-        if _debug: DeviceInfoCache._debug("    - current_info: %r", current_info)
+        if _debug: DeviceInfoCache._debug("    - device_info: %r", device_info)
 
-        return current_info
+        return device_info
 
-    def update_device_info(self, info):
-        """The application has updated one or more fields in the device
-        information record and the cache needs to be updated to reflect the
-        changes.  If this is a cached version of a persistent record then this
-        is the opportunity to update the database."""
-        if _debug: DeviceInfoCache._debug("update_device_info %r", info)
-
-        cache_id, cache_address = info._cache_keys
-
-        if (cache_id is not None) and (info.deviceIdentifier != cache_id):
-            if _debug: DeviceInfoCache._debug("    - device identifier updated")
-
-            # remove the old reference, add the new one
-            del self.cache[cache_id]
-            self.cache[info.deviceIdentifier] = info
-
-            cache_id = info.deviceIdentifier
-
-        if (cache_address is not None) and (info.address != cache_address):
-            if _debug: DeviceInfoCache._debug("    - device address updated")
-
-            # remove the old reference, add the new one
-            del self.cache[cache_address]
-            self.cache[info.address] = info
-
-            cache_address = info.address
-
-        # update the keys
-        info._cache_keys = (cache_id, cache_address)
-
-    def release_device_info(self, info):
+    def release(self, device_info):
         """This function is called by the segmentation state machine when it
         has finished with the device information."""
-        if _debug: DeviceInfoCache._debug("release_device_info %r", info)
+        if _debug: DeviceInfoCache._debug("release %r", device_info)
 
         # this information record might be used by more than one SSM
-        if info._ref_count > 1:
-            if _debug: DeviceInfoCache._debug("    - multiple references")
-            info._ref_count -= 1
-            return
+        if device_info._ref_count == 0:
+            raise RuntimeError("reference count")
 
-        cache_id, cache_address = info._cache_keys
-        if cache_id is not None:
-            del self.cache[cache_id]
-        if cache_address is not None:
-            del self.cache[cache_address]
-        if _debug: DeviceInfoCache._debug("    - released")
+        # decrement the reference count
+        device_info._ref_count -= 1
 
 #
 #   Application
@@ -260,9 +268,9 @@ class Application(ApplicationServiceElement, Collector):
 
         # make sure it hasn't already been defined
         if object_name in self.objectName:
-            raise RuntimeError("already an object with name {0!r}".format(object_name))
+            raise RuntimeError("already an object with name %r" % (object_name,))
         if object_identifier in self.objectIdentifier:
-            raise RuntimeError("already an object with identifier {0!r}".format(object_identifier))
+            raise RuntimeError("already an object with identifier %r" % (object_identifier,))
 
         # now put it in local dictionaries
         self.objectName[object_name] = obj
@@ -470,7 +478,7 @@ class BIPSimpleApplication(ApplicationIOController, WhoIsIAmServices, ReadWriteP
 
     def __init__(self, localDevice, localAddress, deviceInfoCache=None, aseID=None):
         if _debug: BIPSimpleApplication._debug("__init__ %r %r deviceInfoCache=%r aseID=%r", localDevice, localAddress, deviceInfoCache, aseID)
-        ApplicationIOController.__init__(self, localDevice, deviceInfoCache, aseID=aseID)
+        ApplicationIOController.__init__(self, localDevice, localAddress, deviceInfoCache, aseID=aseID)
 
         # local address might be useful for subclasses
         if isinstance(localAddress, Address):
@@ -509,7 +517,7 @@ class BIPSimpleApplication(ApplicationIOController, WhoIsIAmServices, ReadWriteP
         bind(self.bip, self.annexj, self.mux.annexJ)
 
         # bind the BIP stack to the network, no network number
-        self.nsap.bind(self.bip)
+        self.nsap.bind(self.bip, address=self.localAddress)
 
     def close_socket(self):
         if _debug: BIPSimpleApplication._debug("close_socket")
@@ -524,9 +532,14 @@ class BIPSimpleApplication(ApplicationIOController, WhoIsIAmServices, ReadWriteP
 @bacpypes_debugging
 class BIPForeignApplication(ApplicationIOController, WhoIsIAmServices, ReadWritePropertyServices):
 
-    def __init__(self, localDevice, localAddress, bbmdAddress, bbmdTTL, aseID=None):
-        if _debug: BIPForeignApplication._debug("__init__ %r %r %r %r aseID=%r", localDevice, localAddress, bbmdAddress, bbmdTTL, aseID)
-        ApplicationIOController.__init__(self, localDevice, aseID=aseID)
+    def __init__(self, localDevice, localAddress, bbmdAddress, bbmdTTL, deviceInfoCache=None, aseID=None):
+        if _debug:
+            BIPForeignApplication._debug(
+                "__init__ %r %r %r %r deviceInfoCache=%r aseID=%r",
+                localDevice, localAddress, bbmdAddress, bbmdTTL,
+                deviceInfoCache, aseID,
+                )
+        ApplicationIOController.__init__(self, localDevice, localAddress, deviceInfoCache, aseID=aseID)
 
         # local address might be useful for subclasses
         if isinstance(localAddress, Address):
@@ -565,7 +578,7 @@ class BIPForeignApplication(ApplicationIOController, WhoIsIAmServices, ReadWrite
         bind(self.bip, self.annexj, self.mux.annexJ)
 
         # bind the NSAP to the stack, no network number
-        self.nsap.bind(self.bip)
+        self.nsap.bind(self.bip, address=self.localAddress)
 
     def close_socket(self):
         if _debug: BIPForeignApplication._debug("close_socket")
@@ -580,7 +593,7 @@ class BIPForeignApplication(ApplicationIOController, WhoIsIAmServices, ReadWrite
 @bacpypes_debugging
 class BIPNetworkApplication(NetworkServiceElement):
 
-    def __init__(self, localAddress, eID=None):
+    def __init__(self, localAddress, bbmdAddress=None, bbmdTTL=None, eID=None):
         if _debug: BIPNetworkApplication._debug("__init__ %r eID=%r", localAddress, eID)
         NetworkServiceElement.__init__(self, eID)
 
@@ -598,12 +611,16 @@ class BIPNetworkApplication(NetworkServiceElement):
 
         # create a generic BIP stack, bound to the Annex J server
         # on the UDP multiplexer
-        self.bip = BIPSimple()
+        if (not bbmdAddress) and (not bbmdTTL):
+            self.bip = BIPSimple()
+        else:
+            self.bip = BIPForeign(bbmdAddress, bbmdTTL)
         self.annexj = AnnexJCodec()
-        self.mux = UDPMultiplexer(self.localAddress)
+        self.mux = UDPMultiplexer(self.localAddress, noBroadcast=True)
 
         # bind the bottom layers
         bind(self.bip, self.annexj, self.mux.annexJ)
 
         # bind the NSAP to the stack, no network number
-        self.nsap.bind(self.bip)
+        self.nsap.bind(self.bip, address=self.localAddress)
+
