@@ -8,7 +8,7 @@ from .debugging import ModuleLogger
 
 from .primitivedata import BitString, Boolean, CharacterString, Date, Double, \
     Enumerated, Integer, Null, ObjectIdentifier, OctetString, Real, Time, \
-    Unsigned
+    Unsigned, Unsigned16
 from .constructeddata import Any, AnyAtomic, ArrayOf, Choice, Element, \
     Sequence, SequenceOf
 
@@ -94,6 +94,7 @@ class ObjectTypesSupported(BitString):
         , 'accessUser':35
         , 'accessZone':36
         , 'credentialDataInput':37
+        , 'networkPort':56
         , 'networkSecurity':38
         , 'bitstringValue':39
         , 'characterstringValue':40
@@ -107,8 +108,12 @@ class ObjectTypesSupported(BitString):
         , 'positiveIntegerValue':48
         , 'timePatternValue':49
         , 'timeValue':50
+        , 'notificationForwarder':51
+        , 'alertEnrollment':52
+        , 'channel':53
+        , 'lightingOutput':54
         }
-    bitLen = 51
+    bitLen = 55
 
 class ResultFlags(BitString):
     bitNames = \
@@ -1097,9 +1102,50 @@ class ProgramState(Enumerated):
         }
 
 class PropertyIdentifier(Enumerated):
+    # TODO: Sort Alphabetically
     vendor_range = (512, 4194303)
     enumerations = \
         { 'absenteeLimit':244
+        , 'tags':486
+        , 'profileLocation':91
+        , 'eventDetectionEnabled':353
+        , 'apduLength':388
+        , 'linkSpeed':420
+        , 'linkSpeeds':421
+        , 'linkSpeedAutonegotiate':422
+        , 'networkInterfaceName':424
+        , 'bacnetIPMode':408
+        , 'ipAddress':400
+        , 'bacnetIPUDPPort':412
+        , 'ipSubnetMask':411
+        , 'ipDefaultGateway':401
+        , 'bacnetIPMulticastAddress':409
+        , 'ipDNSServer':406
+        , 'ipDHCPEnable':402
+        , 'ipDHCPLeaseTime':403
+        , 'ipDHCPLeaseTimeRemaining':404
+        , 'ipDHCPServer':405
+        , 'bacnetIPNATTraversal':410
+        , 'bacnetIPGlobalAddress':407
+        , 'bbmdBroadcastDistributionTable':414
+        , 'bbmdAcceptFDRegistrations':413
+        , 'bbmdForeignDeviceTable':415
+        , 'fdBBMDAddress':418
+        , 'fdSubscriptionLifetime':419
+        , 'bacnetIPv6Mode':435
+        , 'ipv6Address':436
+        , 'ipv6PrefixLength':437
+        , 'bacnetIPv6UDPPort':438
+        , 'ipv6DefaultGateway':439
+        , 'bacnetIPv6MulticastAddress':440
+        , 'ipv6DNSServer':441
+        , 'ipv6AutoAddressingEnabled':442
+        , 'ipv6DHCPLeaseTime':443
+        , 'ipv6DHCPLeaseTimeRemaining':444
+        , 'ipv6DHCPServer':445
+        , 'ipv6ZoneIndex':446
+        , 'virtualMACAddressTable':429
+        , 'routingTable':428
         , 'acceptedModes':175
         , 'accessAlarmEvents':245
         , 'accessDoors':246
@@ -1154,8 +1200,10 @@ class PropertyIdentifier(Enumerated):
         , 'bufferSize':126
         , 'changeOfStateCount':15
         , 'changeOfStateTime':16
+        , 'changesPending':416
         , 'channelNumber':366
         , 'clientCovIncrement':127
+        , 'command':417
         , 'configurationFiles':154
         , 'controlGroups':367
         , 'controlledVariableReference':19
@@ -1282,6 +1330,7 @@ class PropertyIdentifier(Enumerated):
         , 'loggingRecord':184
         , 'loggingType':197
         , 'lowLimit':59
+        , 'macAddress':423
         , 'maintenanceRequired':158
         , 'manipulatedVariableReference':60
         , 'manualSlaveAddressBinding':170
@@ -1313,6 +1362,9 @@ class PropertyIdentifier(Enumerated):
         , 'musterPoint':287
         , 'negativeAccessRules':288
         , 'networkAccessSecurityPolicies':332
+        , 'networkNumber':425
+        , 'networkNumberQuality':427
+        , 'networkType': 427
         , 'nodeSubtype':207
         , 'nodeType':208
         , 'notificationClass':17
@@ -1361,6 +1413,7 @@ class PropertyIdentifier(Enumerated):
         , 'propertyList':371
         , 'proportionalConstant':93
         , 'proportionalConstantUnits':94
+        , 'protocolLevel':482
         , 'protocolObjectTypesSupported':96
         , 'protocolRevision':139
         , 'protocolServicesSupported':97
@@ -1372,6 +1425,7 @@ class PropertyIdentifier(Enumerated):
         , 'recipientList':102
         , 'recordsSinceNotification':140
         , 'recordCount':141
+        , 'referencePort':483
         , 'reliability':103
         , 'reliabilityEvaluationInhibit':357
         , 'relinquishDefault':104
@@ -1562,10 +1616,112 @@ class WriteStatus(Enumerated):
         , 'failed':3
         }
 
+class NetworkType(Enumerated):
+    enumerations = \
+        { 'ethernet':0
+        , 'arcnet':1
+        , 'mstp':2
+        , 'ptp':3
+        , 'lontalk':4
+        , 'ipv4':5
+        , 'zigbee':6
+        , 'virtual': 7
+        # , 'non-bacnet': 8  Removed in Version 1, Revision 18
+        , 'ipv6':9
+        , 'serial':10
+        }
+
+class ProtocolLevel(Enumerated):
+    enumerations = \
+        { 'physical':0
+        , 'protocol':1
+        , 'bacnetApplication':2
+        , 'nonBacnetApplication':3
+        }
+
+class NetworkNumberQuality(Enumerated):
+    enumerations = \
+        { 'unknown':0
+        , 'learned':1
+        , 'learnedConfigured':2
+        , 'configured':3
+        }
+
+class NetworkPortCommand(Enumerated):
+    enumerations = \
+        { 'idle':0
+        , 'discardChanges':1
+        , 'renewFdDRegistration':2
+        , 'restartSlaveDiscovery':3
+        , 'renewDHCP':4
+        , 'restartAutonegotiation':5
+        , 'disconnect':6
+        , 'restartPort':7
+        }
+
+class IPMode(Enumerated):
+    enumerations = \
+        { 'normal':0
+        , 'foreign':1
+        , 'bbmd':2
+        }
+
+class RouterEntryStatus(Enumerated):
+    enumerations = \
+        { 'available':0
+        , 'busy':1
+        , 'disconnected':2
+        }
+
 #
 #   Forward Sequences
 #
 
+class HostAddress(Choice):
+    choiceElements = \
+        [ Element('none', Null)
+        , Element('ipAddress', OctetString)  # 4 octets for B/IP or 16 octets for B/IPv6
+        , Element('name', CharacterString)  # Internet host name (see RFC 1123)
+        ]
+
+class HostNPort(Sequence):
+    sequenceElements = \
+        [ Element('host', HostAddress)
+        , Element('port', Unsigned16)
+        ]
+
+class BDTEntry(Sequence):
+    sequenceElements = \
+        [ Element('bbmdAddress', HostNPort)
+        , Element('broadcastMask', OctetString)  # shall be present if BACnet/IP, and absent for BACnet/IPv6
+        ]
+
+class FDTEntry(Sequence):
+    sequenceElements = \
+        [ Element('bacnetIPAddress', OctetString)  # the 6-octet B/IP or 18-octet B/IPv6 address of the registrant
+        , Element('timeToLive', Unsigned16)  # time to live in seconds at the time of registration
+        , Element('remainingTimeToLive', Unsigned16)  # remaining time to live in seconds, incl. grace period
+        ]
+
+class VMACEntry(Sequence):
+    sequenceElements = \
+        [ Element('virtualMACAddress', OctetString)  # maximum size 6 octets
+        , Element('nativeMACAddress', OctetString)
+        ]
+
+class RouterEntry(Sequence):
+    sequenceElements = \
+        [ Element('networkNumber', Unsigned16)
+        , Element('macAddress', OctetString)
+        , Element('status', RouterEntryStatus)  # Defined Above
+        ]
+
+class NameValue(Sequence):
+    sequenceElements = \
+        [ Element('name', CharacterString)
+        , Element('value', AnyAtomic)  # IS ATOMIC CORRECT HERE? value is limited to primitive datatypes and BACnetDateTime
+        ]
+
 class DeviceAddress(Sequence):
     sequenceElements = \
         [ Element('networkNumber', Unsigned)
@@ -1604,6 +1760,16 @@ class ErrorType(Sequence):
         , Element('errorCode', ErrorCode)
         ]
 
+class LightingCommand(Sequence):
+    sequenceElements = \
+        [ Element('operation', LightingOperation, 0)
+        , Element('targetLevel', Real, 1, True)
+        , Element('rampRate', Real, 2, True)
+        , Element('stepIncrement', Real, 3, True)
+        , Element('fadeTime', Unsigned, 4, True)
+        , Element('priority', Unsigned, 5, True)
+        ]
+
 class ObjectPropertyReference(Sequence):
     sequenceElements = \
         [ Element('objectIdentifier', ObjectIdentifier, 0)
@@ -1716,7 +1882,8 @@ class AccessRule(Sequence):
         ]
 
 class AccessThreatLevel(Unsigned):
-    pass
+    _low_limit = 0
+    _high_limit = 100
 
 class AccumulatorRecord(Sequence):
     sequenceElements = \
@@ -1792,8 +1959,21 @@ class CalendarEntry(Choice):
         ]
 
 class ChannelValue(Choice):
-    choiceElements = [
-        ### needs help
+    choiceElements = \
+        [ Element('null', Null)
+        , Element('real', Real)
+        , Element('enumerated', Enumerated)
+        , Element('unsigned', Unsigned)
+        , Element('boolean', Boolean)
+        , Element('integer', Integer)
+        , Element('double', Double)
+        , Element('time', Time)
+        , Element('characterString', CharacterString)
+        , Element('octetString', OctetString)
+        , Element('bitString', BitString)
+        , Element('date', Date)
+        , Element('objectidentifier', ObjectIdentifier)
+        , Element('lightingCommand', LightingCommand, 0)
         ]
 
 class ClientCOV(Choice):
@@ -2068,16 +2248,6 @@ class KeyIdentifier(Sequence):
         , Element('keyId', Unsigned, 1)
         ]
 
-class LightingCommand(Sequence):
-    sequenceElements = \
-        [ Element('operation', LightingOperation, 0)
-        , Element('targetLevel', Real, 1) ### optional
-        , Element('rampRate', Real, 2)   ### optional
-        , Element('stepIncrement', Real, 3) ### optional
-        , Element('fadeTime', Unsigned, 4) ### optional
-        , Element('priority', Unsigned, 5) ### optional
-        ]
-
 class LogDataLogData(Choice):
     choiceElements = \
         [ Element('booleanValue', Boolean, 0)
@@ -2336,19 +2506,21 @@ class PriorityValue(Choice):
         , Element('enumerated', Enumerated)
         , Element('unsigned', Unsigned)
         , Element('boolean', Boolean)
-        , Element('signed', Integer)
+        , Element('integer', Integer)
         , Element('double', Double)
         , Element('time', Time)
         , Element('characterString', CharacterString)
         , Element('octetString', OctetString)
         , Element('bitString', BitString)
         , Element('date', Date)
-        , Element('objectid', ObjectIdentifier)
+        , Element('objectidentifier', ObjectIdentifier)
         , Element('constructedValue', Any, 0)
         , Element('datetime', DateTime, 1)
         ]
 
-class PriorityArray(ArrayOf(PriorityValue)):
+class PriorityArray(ArrayOf(
+        PriorityValue, fixed_length=16, prototype=PriorityValue(null=()),
+        )):
     pass
 
 class PropertyAccessResultAccessResult(Choice):
