@@ -15,7 +15,7 @@ from .task import OneShotTask, RecurringTask
 from .comm import Client, Server, bind, \
     ServiceAccessPoint, ApplicationServiceElement
 
-from .pdu import Address, LocalBroadcast, LocalStation, PDU, \
+from .pdu import Address, LocalBroadcast, PDU, \
     unpack_ip_addr
 from .bvll import BVLPDU, DeleteForeignDeviceTableEntry, \
     DistributeBroadcastToNetwork, FDTEntry, ForwardedNPDU, \
@@ -404,6 +404,54 @@ class BIPSimple(BIPSAP, Client, Server):
             # send it upstream
             self.response(xpdu)
 
+        elif isinstance(pdu, WriteBroadcastDistributionTable):
+            # build a response
+            xpdu = Result(code=0x0010, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadBroadcastDistributionTable):
+            # build a response
+            xpdu = Result(code=0x0020, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, RegisterForeignDevice):
+            # build a response
+            xpdu = Result(code=0x0030, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadForeignDeviceTable):
+            # build a response
+            xpdu = Result(code=0x0040, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, DeleteForeignDeviceTableEntry):
+            # build a response
+            xpdu = Result(code=0x0050, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, DistributeBroadcastToNetwork):
+            # build a response
+            xpdu = Result(code=0x0060, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
         else:
             BIPSimple._warning("invalid pdu type: %s", type(pdu))
 
@@ -442,11 +490,6 @@ class BIPForeign(BIPSAP, Client, Server, OneShotTask, DebugContents):
     def indication(self, pdu):
         if _debug: BIPForeign._debug("indication %r", pdu)
 
-        # check the BBMD registration status, we may not be registered
-        if self.registrationStatus != 0:
-            if _debug: BIPForeign._debug("    - packet dropped, unregistered")
-            return
-
         # check for local stations
         if pdu.pduDestination.addrType == Address.localStationAddr:
             # make an original unicast PDU
@@ -458,6 +501,11 @@ class BIPForeign(BIPSAP, Client, Server, OneShotTask, DebugContents):
 
         # check for broadcasts
         elif pdu.pduDestination.addrType == Address.localBroadcastAddr:
+            # check the BBMD registration status, we may not be registered
+            if self.registrationStatus != 0:
+                if _debug: BIPForeign._debug("    - packet dropped, unregistered")
+                return
+
             # make an original broadcast PDU
             xpdu = DistributeBroadcastToNetwork(pdu, user_data=pdu.pduUserData)
             xpdu.pduDestination = self.bbmdAddress
@@ -494,20 +542,7 @@ class BIPForeign(BIPSAP, Client, Server, OneShotTask, DebugContents):
 
             return
 
-        # check the BBMD registration status, we may not be registered
-        if self.registrationStatus != 0:
-            if _debug: BIPForeign._debug("    - packet dropped, unregistered")
-            return
-
-        if isinstance(pdu, ReadBroadcastDistributionTableAck):
-            # send this to the service access point
-            self.sap_response(pdu)
-
-        elif isinstance(pdu, ReadForeignDeviceTableAck):
-            # send this to the service access point
-            self.sap_response(pdu)
-
-        elif isinstance(pdu, OriginalUnicastNPDU):
+        if isinstance(pdu, OriginalUnicastNPDU):
             # build a vanilla PDU
             xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=pdu.pduDestination, user_data=pdu.pduUserData)
 
@@ -515,12 +550,81 @@ class BIPForeign(BIPSAP, Client, Server, OneShotTask, DebugContents):
             self.response(xpdu)
 
         elif isinstance(pdu, ForwardedNPDU):
+            # check the BBMD registration status, we may not be registered
+            if self.registrationStatus != 0:
+                if _debug: BIPForeign._debug("    - packet dropped, unregistered")
+                return
+
+            # make sure the forwarded PDU from the bbmd
+            if pdu.pduSource != self.bbmdAddress:
+                if _debug: BIPForeign._debug("    - packet dropped, not from the BBMD")
+                return
+
             # build a PDU with the source from the real source
             xpdu = PDU(pdu.pduData, source=pdu.bvlciAddress, destination=LocalBroadcast(), user_data=pdu.pduUserData)
 
             # send it upstream
             self.response(xpdu)
 
+        elif isinstance(pdu, ReadBroadcastDistributionTableAck):
+            # send this to the service access point
+            self.sap_response(pdu)
+
+        elif isinstance(pdu, ReadForeignDeviceTableAck):
+            # send this to the service access point
+            self.sap_response(pdu)
+
+        elif isinstance(pdu, WriteBroadcastDistributionTable):
+            # build a response
+            xpdu = Result(code=0x0010, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadBroadcastDistributionTable):
+            # build a response
+            xpdu = Result(code=0x0020, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, RegisterForeignDevice):
+            # build a response
+            xpdu = Result(code=0x0030, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadForeignDeviceTable):
+            # build a response
+            xpdu = Result(code=0x0040, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, DeleteForeignDeviceTableEntry):
+            # build a response
+            xpdu = Result(code=0x0050, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, DistributeBroadcastToNetwork):
+            # build a response
+            xpdu = Result(code=0x0060, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, OriginalBroadcastNPDU):
+            if _debug: BIPForeign._debug("    - packet dropped")
+
         else:
             BIPForeign._warning("invalid pdu type: %s", type(pdu))
 
@@ -538,7 +642,7 @@ class BIPForeign(BIPSAP, Client, Server, OneShotTask, DebugContents):
         self.bbmdTimeToLive = ttl
 
         # install this task to run when it gets a chance
-        self.install_task(delta=0)
+        self.install_task(when=0)
 
     def unregister(self):
         """Drop the registration with a BBMD."""
@@ -595,7 +699,7 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
             # make an original unicast PDU
             xpdu = OriginalUnicastNPDU(pdu, user_data=pdu.pduUserData)
             xpdu.pduDestination = pdu.pduDestination
-            if _debug: BIPBBMD._debug("    - xpdu: %r", xpdu)
+            if _debug: BIPBBMD._debug("    - original unicast xpdu: %r", xpdu)
 
             # send it downstream
             self.request(xpdu)
@@ -618,13 +722,13 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
             for bdte in self.bbmdBDT:
                 if bdte != self.bbmdAddress:
                     xpdu.pduDestination = Address( ((bdte.addrIP|~bdte.addrMask), bdte.addrPort) )
-                    BIPBBMD._debug("        - sending to peer: %r", xpdu.pduDestination)
+                    BIPBBMD._debug("    - sending to peer: %r", xpdu.pduDestination)
                     self.request(xpdu)
 
             # send it to the registered foreign devices
             for fdte in self.bbmdFDT:
                 xpdu.pduDestination = fdte.fdAddress
-                if _debug: BIPBBMD._debug("        - sending to foreign device: %r", xpdu.pduDestination)
+                if _debug: BIPBBMD._debug("    - sending to foreign device: %r", xpdu.pduDestination)
                 self.request(xpdu)
 
         else:
@@ -640,8 +744,9 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
 
         elif isinstance(pdu, WriteBroadcastDistributionTable):
             # build a response
-            xpdu = Result(code=99, user_data=pdu.pduUserData)
+            xpdu = Result(code=0x0010, user_data=pdu.pduUserData)
             xpdu.pduDestination = pdu.pduSource
+            if _debug: BIPBBMD._debug("    - xpdu: %r", xpdu)
 
             # send it downstream
             self.request(xpdu)
@@ -660,27 +765,38 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
             self.sap_response(pdu)
 
         elif isinstance(pdu, ForwardedNPDU):
-            # build a PDU with the source from the real source
-            xpdu = PDU(pdu.pduData, source=pdu.bvlciAddress, destination=LocalBroadcast(), user_data=pdu.pduUserData)
-            if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
+            # send it upstream if there is a network layer
+            if self.serverPeer:
+                # build a PDU with a local broadcast address
+                xpdu = PDU(pdu.pduData, source=pdu.bvlciAddress, destination=LocalBroadcast(), user_data=pdu.pduUserData)
+                if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
 
-            # send it upstream
-            self.response(xpdu)
+                self.response(xpdu)
 
             # build a forwarded NPDU to send out
             xpdu = ForwardedNPDU(pdu.bvlciAddress, pdu, destination=None, user_data=pdu.pduUserData)
             if _debug: BIPBBMD._debug("    - forwarded xpdu: %r", xpdu)
 
-            # look for self as first entry in the BDT
-            if self.bbmdBDT and (self.bbmdBDT[0] == self.bbmdAddress):
-                xpdu.pduDestination = LocalBroadcast()
-                if _debug: BIPBBMD._debug("        - local broadcast")
-                self.request(xpdu)
+            # if this was unicast to us, do next hop
+            if pdu.pduDestination.addrType == Address.localStationAddr:
+                if _debug: BIPBBMD._debug("    - unicast message")
+
+                # if this BBMD is listed in its BDT, send a local broadcast
+                if self.bbmdAddress in self.bbmdBDT:
+                    xpdu.pduDestination = LocalBroadcast()
+                    if _debug: BIPBBMD._debug("    - local broadcast")
+                    self.request(xpdu)
+
+            elif pdu.pduDestination.addrType == Address.localBroadcastAddr:
+                if _debug: BIPBBMD._debug("    - directed broadcast message")
+
+            else:
+                BIPBBMD._warning("invalid destination address: %r", pdu.pduDestination)
 
             # send it to the registered foreign devices
             for fdte in self.bbmdFDT:
                 xpdu.pduDestination = fdte.fdAddress
-                if _debug: BIPBBMD._debug("        - sending to foreign device: %r", xpdu.pduDestination)
+                if _debug: BIPBBMD._debug("    - sending to foreign device: %r", xpdu.pduDestination)
                 self.request(xpdu)
 
         elif isinstance(pdu, RegisterForeignDevice):
@@ -719,12 +835,13 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
             self.request(xpdu)
 
         elif isinstance(pdu, DistributeBroadcastToNetwork):
-            # build a PDU with a local broadcast address
-            xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=LocalBroadcast(), user_data=pdu.pduUserData)
-            if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
+            # send it upstream if there is a network layer
+            if self.serverPeer:
+                # build a PDU with a local broadcast address
+                xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=LocalBroadcast(), user_data=pdu.pduUserData)
+                if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
 
-            # send it upstream
-            self.response(xpdu)
+                self.response(xpdu)
 
             # build a forwarded NPDU to send out
             xpdu = ForwardedNPDU(pdu.pduSource, pdu, user_data=pdu.pduUserData)
@@ -734,35 +851,37 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
             for bdte in self.bbmdBDT:
                 if bdte == self.bbmdAddress:
                     xpdu.pduDestination = LocalBroadcast()
-                    if _debug: BIPBBMD._debug("        - local broadcast")
+                    if _debug: BIPBBMD._debug("    - local broadcast")
                     self.request(xpdu)
                 else:
                     xpdu.pduDestination = Address( ((bdte.addrIP|~bdte.addrMask), bdte.addrPort) )
-                    if _debug: BIPBBMD._debug("        - sending to peer: %r", xpdu.pduDestination)
+                    if _debug: BIPBBMD._debug("    - sending to peer: %r", xpdu.pduDestination)
                     self.request(xpdu)
 
             # send it to the other registered foreign devices
             for fdte in self.bbmdFDT:
                 if fdte.fdAddress != pdu.pduSource:
                     xpdu.pduDestination = fdte.fdAddress
-                    if _debug: BIPBBMD._debug("        - sending to foreign device: %r", xpdu.pduDestination)
+                    if _debug: BIPBBMD._debug("    - sending to foreign device: %r", xpdu.pduDestination)
                     self.request(xpdu)
 
         elif isinstance(pdu, OriginalUnicastNPDU):
-            # build a vanilla PDU
-            xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=pdu.pduDestination, user_data=pdu.pduUserData)
-            if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
+            # send it upstream if there is a network layer
+            if self.serverPeer:
+                # build a PDU with a local broadcast address
+                xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=pdu.pduDestination, user_data=pdu.pduUserData)
+                if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
 
-            # send it upstream
-            self.response(xpdu)
+                self.response(xpdu)
 
         elif isinstance(pdu, OriginalBroadcastNPDU):
-            # build a PDU with a local broadcast address
-            xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=LocalBroadcast(), user_data=pdu.pduUserData)
-            if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
+            # send it upstream if there is a network layer
+            if self.serverPeer:
+                # build a PDU with a local broadcast address
+                xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=LocalBroadcast(), user_data=pdu.pduUserData)
+                if _debug: BIPBBMD._debug("    - upstream xpdu: %r", xpdu)
 
-            # send it upstream
-            self.response(xpdu)
+                self.response(xpdu)
 
             # make a forwarded PDU
             xpdu = ForwardedNPDU(pdu.pduSource, pdu, user_data=pdu.pduUserData)
@@ -772,13 +891,13 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
             for bdte in self.bbmdBDT:
                 if bdte != self.bbmdAddress:
                     xpdu.pduDestination = Address( ((bdte.addrIP|~bdte.addrMask), bdte.addrPort) )
-                    if _debug: BIPBBMD._debug("        - sending to peer: %r", xpdu.pduDestination)
+                    if _debug: BIPBBMD._debug("    - sending to peer: %r", xpdu.pduDestination)
                     self.request(xpdu)
 
             # send it to the registered foreign devices
             for fdte in self.bbmdFDT:
                 xpdu.pduDestination = fdte.fdAddress
-                if _debug: BIPBBMD._debug("        - sending to foreign device: %r", xpdu.pduDestination)
+                if _debug: BIPBBMD._debug("    - sending to foreign device: %r", xpdu.pduDestination)
                 self.request(xpdu)
 
         else:
@@ -792,7 +911,7 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
         if isinstance(addr, Address):
             pass
         elif isinstance(addr, str):
-            addr = LocalStation( addr )
+            addr = Address(addr)
         else:
             raise TypeError("addr must be a string or an Address")
 
@@ -817,7 +936,305 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
         if isinstance(addr, Address):
             pass
         elif isinstance(addr, str):
-            addr = LocalStation( addr )
+            addr = Address(addr)
+        else:
+            raise TypeError("addr must be a string or an Address")
+
+        # find it and delete it
+        stat = 0
+        for i in range(len(self.bbmdFDT)-1, -1, -1):
+            if addr == self.bbmdFDT[i].fdAddress:
+                del self.bbmdFDT[i]
+                break
+        else:
+            stat = 0x0050 ### entry not found
+
+        # return status
+        return stat
+
+    def process_task(self):
+        # look for foreign device registrations that have expired
+        for i in range(len(self.bbmdFDT)-1, -1, -1):
+            fdte = self.bbmdFDT[i]
+            fdte.fdRemain -= 1
+
+            # delete it if it expired
+            if fdte.fdRemain <= 0:
+                if _debug: BIPBBMD._debug("foreign device expired: %r", fdte.fdAddress)
+                del self.bbmdFDT[i]
+
+    def add_peer(self, addr):
+        if _debug: BIPBBMD._debug("add_peer %r", addr)
+
+        # see if it is an address or make it one
+        if isinstance(addr, Address):
+            pass
+        elif isinstance(addr, str):
+            addr = Address(addr)
+        else:
+            raise TypeError("addr must be a string or an Address")
+
+        # see if it's already there
+        for bdte in self.bbmdBDT:
+            if addr == bdte:
+                break
+        else:
+            self.bbmdBDT.append(addr)
+
+    def delete_peer(self, addr):
+        if _debug: BIPBBMD._debug("delete_peer %r", addr)
+
+        # see if it is an address or make it one
+        if isinstance(addr, Address):
+            pass
+        elif isinstance(addr, str):
+            addr = Address(addr)
+        else:
+            raise TypeError("addr must be a string or an Address")
+
+        # look for the peer address
+        for i in range(len(self.bbmdBDT)-1, -1, -1):
+            if addr == self.bbmdBDT[i]:
+                del self.bbmdBDT[i]
+                break
+        else:
+            pass
+
+#
+#   BIPNAT
+#
+
+@bacpypes_debugging
+class BIPNAT(BIPSAP, Client, Server, RecurringTask, DebugContents):
+
+    _debug_contents = ('bbmdAddress', 'bbmdBDT+', 'bbmdFDT+')
+
+    def __init__(self, addr, sapID=None, cid=None, sid=None):
+        """A BBMD node that is the destination for NATed traffic."""
+        if _debug: BIPNAT._debug("__init__ %r sapID=%r cid=%r sid=%r", addr, sapID, cid, sid)
+        BIPSAP.__init__(self, sapID)
+        Client.__init__(self, cid)
+        Server.__init__(self, sid)
+        RecurringTask.__init__(self, 1000.0)
+
+        self.bbmdAddress = addr
+        self.bbmdBDT = []
+        self.bbmdFDT = []
+
+        # install so process_task runs
+        self.install_task()
+
+    def indication(self, pdu):
+        if _debug: BIPNAT._debug("indication %r", pdu)
+
+        # check for local stations
+        if pdu.pduDestination.addrType == Address.localStationAddr:
+            ###TODO the destination should be a peer or a registered foreign device
+
+            # make an original unicast PDU
+            xpdu = OriginalUnicastNPDU(pdu, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduDestination
+            if _debug: BIPNAT._debug("    - xpdu: %r", xpdu)
+
+            # send it downstream
+            self.request(xpdu)
+
+        # check for broadcasts
+        elif pdu.pduDestination.addrType == Address.localBroadcastAddr:
+            # make a forwarded PDU
+            xpdu = ForwardedNPDU(self.bbmdAddress, pdu, user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - forwarded xpdu: %r", xpdu)
+
+            # send it to the peers, all of them have all F's mask
+            for bdte in self.bbmdBDT:
+                if bdte != self.bbmdAddress:
+                    xpdu.pduDestination = Address((bdte.addrIP, bdte.addrPort))
+                    BIPNAT._debug("        - sending to peer: %r", xpdu.pduDestination)
+                    self.request(xpdu)
+
+            # send it to the registered foreign devices
+            for fdte in self.bbmdFDT:
+                xpdu.pduDestination = fdte.fdAddress
+                if _debug: BIPNAT._debug("        - sending to foreign device: %r", xpdu.pduDestination)
+                self.request(xpdu)
+
+        else:
+            BIPNAT._warning("invalid destination address: %r", pdu.pduDestination)
+
+    def confirmation(self, pdu):
+        if _debug: BIPNAT._debug("confirmation %r",  pdu)
+
+        # some kind of response to a request
+        if isinstance(pdu, Result):
+            # send this to the service access point
+            self.sap_response(pdu)
+
+        elif isinstance(pdu, WriteBroadcastDistributionTable):
+            ###TODO verify this is from a management network/address
+
+            # build a response
+            xpdu = Result(code=99, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadBroadcastDistributionTable):
+            ###TODO verify this is from a management network/address
+
+            # build a response
+            xpdu = ReadBroadcastDistributionTableAck(self.bbmdBDT, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+            if _debug: BIPNAT._debug("    - xpdu: %r", xpdu)
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadBroadcastDistributionTableAck):
+            # send this to the service access point
+            self.sap_response(pdu)
+
+        elif isinstance(pdu, ForwardedNPDU):
+            ###TODO verify this is from a peer
+
+            # build a PDU with the source from the real source
+            xpdu = PDU(pdu.pduData, source=pdu.bvlciAddress, destination=LocalBroadcast(), user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - upstream xpdu: %r", xpdu)
+
+            # send it upstream
+            self.response(xpdu)
+
+            # build a forwarded NPDU to send out
+            xpdu = ForwardedNPDU(pdu.bvlciAddress, pdu, destination=None, user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - forwarded xpdu: %r", xpdu)
+
+            # send it to the registered foreign devices
+            for fdte in self.bbmdFDT:
+                xpdu.pduDestination = fdte.fdAddress
+                if _debug: BIPNAT._debug("        - sending to foreign device: %r", xpdu.pduDestination)
+                self.request(xpdu)
+
+        elif isinstance(pdu, RegisterForeignDevice):
+            ###TODO verify this is from an acceptable address
+
+            # process the request
+            stat = self.register_foreign_device(pdu.pduSource, pdu.bvlciTimeToLive)
+
+            # build a response
+            xpdu = Result(code=stat, destination=pdu.pduSource, user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - xpdu: %r", xpdu)
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadForeignDeviceTable):
+            ###TODO verify this is from a management network/address
+
+            # build a response
+            xpdu = ReadForeignDeviceTableAck(self.bbmdFDT, destination=pdu.pduSource, user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - xpdu: %r", xpdu)
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, ReadForeignDeviceTableAck):
+            # send this to the service access point
+            self.sap_response(pdu)
+
+        elif isinstance(pdu, DeleteForeignDeviceTableEntry):
+            ###TODO verify this is from a management network/address
+
+            # process the request
+            stat = self.delete_foreign_device_table_entry(pdu.bvlciAddress)
+
+            # build a response
+            xpdu = Result(code=stat, user_data=pdu.pduUserData)
+            xpdu.pduDestination = pdu.pduSource
+            if _debug: BIPNAT._debug("    - xpdu: %r", xpdu)
+
+            # send it downstream
+            self.request(xpdu)
+
+        elif isinstance(pdu, DistributeBroadcastToNetwork):
+            ###TODO verify this is from a registered foreign device
+
+            # build a PDU with a local broadcast address
+            xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=LocalBroadcast(), user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - upstream xpdu: %r", xpdu)
+
+            # send it upstream
+            self.response(xpdu)
+
+            # build a forwarded NPDU to send out
+            xpdu = ForwardedNPDU(pdu.pduSource, pdu, user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - forwarded xpdu: %r", xpdu)
+
+            # send it to the peers
+            for bdte in self.bbmdBDT:
+                if bdte == self.bbmdAddress:
+                    if _debug: BIPNAT._debug("        - no local broadcast")
+                else:
+                    xpdu.pduDestination = Address((bdte.addrIP, bdte.addrPort))
+                    if _debug: BIPNAT._debug("        - sending to peer: %r", xpdu.pduDestination)
+                    self.request(xpdu)
+
+            # send it to the other registered foreign devices
+            for fdte in self.bbmdFDT:
+                if fdte.fdAddress != pdu.pduSource:
+                    xpdu.pduDestination = fdte.fdAddress
+                    if _debug: BIPNAT._debug("        - sending to foreign device: %r", xpdu.pduDestination)
+                    self.request(xpdu)
+
+        elif isinstance(pdu, OriginalUnicastNPDU):
+            ###TODO verify this is from a peer
+
+            # build a vanilla PDU
+            xpdu = PDU(pdu.pduData, source=pdu.pduSource, destination=pdu.pduDestination, user_data=pdu.pduUserData)
+            if _debug: BIPNAT._debug("    - upstream xpdu: %r", xpdu)
+
+            # send it upstream
+            self.response(xpdu)
+
+        elif isinstance(pdu, OriginalBroadcastNPDU):
+            if _debug: BIPNAT._debug("    - original broadcast dropped")
+
+        else:
+            BIPNAT._warning("invalid pdu type: %s", type(pdu))
+
+    def register_foreign_device(self, addr, ttl):
+        """Add a foreign device to the FDT."""
+        if _debug: BIPNAT._debug("register_foreign_device %r %r", addr, ttl)
+
+        # see if it is an address or make it one
+        if isinstance(addr, Address):
+            pass
+        elif isinstance(addr, str):
+            addr = Address(addr)
+        else:
+            raise TypeError("addr must be a string or an Address")
+
+        for fdte in self.bbmdFDT:
+            if addr == fdte.fdAddress:
+                break
+        else:
+            fdte = FDTEntry()
+            fdte.fdAddress = addr
+            self.bbmdFDT.append( fdte )
+
+        fdte.fdTTL = ttl
+        fdte.fdRemain = ttl + 5
+
+        # return success
+        return 0
+
+    def delete_foreign_device_table_entry(self, addr):
+        if _debug: BIPNAT._debug("delete_foreign_device_table_entry %r", addr)
+
+        # see if it is an address or make it one
+        if isinstance(addr, Address):
+            pass
+        elif isinstance(addr, str):
+            addr = Address(addr)
         else:
             raise TypeError("addr must be a string or an Address")
 
@@ -841,17 +1258,17 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
 
             # delete it if it expired
             if fdte.fdRemain <= 0:
-                if _debug: BIPBBMD._debug("foreign device expired: %r", fdte.fdAddress)
+                if _debug: BIPNAT._debug("foreign device expired: %r", fdte.fdAddress)
                 del self.bbmdFDT[i]
 
     def add_peer(self, addr):
-        if _debug: BIPBBMD._debug("add_peer %r", addr)
+        if _debug: BIPNAT._debug("add_peer %r", addr)
 
         # see if it is an address or make it one
         if isinstance(addr, Address):
             pass
         elif isinstance(addr, str):
-            addr = LocalStation(addr)
+            addr = Address(addr)
         else:
             raise TypeError("addr must be a string or an Address")
 
@@ -867,13 +1284,13 @@ class BIPBBMD(BIPSAP, Client, Server, RecurringTask, DebugContents):
             self.bbmdBDT.append(addr)
 
     def delete_peer(self, addr):
-        if _debug: BIPBBMD._debug("delete_peer %r", addr)
+        if _debug: BIPNAT._debug("delete_peer %r", addr)
 
         # see if it is an address or make it one
         if isinstance(addr, Address):
             pass
         elif isinstance(addr, str):
-            addr = LocalStation(addr)
+            addr = Address(addr)
         else:
             raise TypeError("addr must be a string or an Address")
 
