@@ -5,8 +5,9 @@ Constructed Data
 """
 
 import sys
+from copy import deepcopy as _deepcopy
 
-from .errors import DecodingError, \
+from .errors import DecodingError, EncodingError, \
     MissingRequiredParameter, InvalidParameterDatatype, InvalidTag
 from .debugging import ModuleLogger, bacpypes_debugging
 
@@ -78,7 +79,7 @@ class Sequence(object):
         """
         """
         if _debug: Sequence._debug("encode %r", taglist)
-        global _sequence_of_classes
+        global _sequence_of_classes, _list_of_classes
 
         # make sure we're dealing with a tag list
         if not isinstance(taglist, TagList):
@@ -90,7 +91,7 @@ class Sequence(object):
                 continue
             if not element.optional and value is None:
                 raise MissingRequiredParameter("%s is a missing required element of %s" % (element.name, self.__class__.__name__))
-            if element.klass in _sequence_of_classes:
+            if (element.klass in _sequence_of_classes) or (element.klass in _list_of_classes):
                 # might need to encode an opening tag
                 if element.context is not None:
                     taglist.append(OpeningTag(element.context))
@@ -134,7 +135,10 @@ class Sequence(object):
                 raise TypeError("%s must be of type %s" % (element.name, element.klass.__name__))
 
     def decode(self, taglist):
+        """
+        """
         if _debug: Sequence._debug("decode %r", taglist)
+        global _sequence_of_classes, _list_of_classes
 
         # make sure we're dealing with a tag list
         if not isinstance(taglist, TagList):
@@ -142,13 +146,14 @@ class Sequence(object):
 
         for element in self.sequenceElements:
             tag = taglist.Peek()
+            if _debug: Sequence._debug("    - element, tag: %r, %r", element, tag)
 
             # no more elements
             if tag is None:
                 if element.optional:
                     # omitted optional element
                     setattr(self, element.name, None)
-                elif element.klass in _sequence_of_classes:
+                elif (element.klass in _sequence_of_classes) or (element.klass in _list_of_classes):
                     # empty list
                     setattr(self, element.name, [])
                 else:
@@ -188,7 +193,29 @@ class Sequence(object):
                     if tag.tagClass != Tag.closingTagClass or tag.tagNumber != element.context:
                         raise InvalidTag("%s expected closing tag %d" % (element.name, element.context))
 
-            # check for an atomic element
+            # check for an any atomic element
+            elif issubclass(element.klass, AnyAtomic):
+                # convert it to application encoding
+                if element.context is not None:
+                    raise InvalidTag("%s any atomic with context tag %d" % (element.name, element.context))
+
+                if tag.tagClass != Tag.applicationTagClass:
+                    if not element.optional:
+                        raise InvalidParameterDatatype("%s expected any atomic application tag" % (element.name,))
+                    else:
+                        setattr(self, element.name, None)
+                        continue
+
+                # consume the tag
+                taglist.Pop()
+
+                # a helper cooperates between the atomic value and the tag
+                helper = element.klass(tag)
+
+                # now save the value
+                setattr(self, element.name, helper.value)
+
+            # check for specific kind of atomic element, or the context says what kind
             elif issubclass(element.klass, Atomic):
                 # convert it to application encoding
                 if element.context is not None:
@@ -285,7 +312,7 @@ class Sequence(object):
                         raise InvalidTag("%s expected closing tag %d" % (element.name, element.context))
 
     def debug_contents(self, indent=1, file=sys.stdout, _ids=None):
-        global _sequence_of_classes
+        global _sequence_of_classes, _list_of_classes
 
         for element in self.sequenceElements:
             value = getattr(self, element.name, None)
@@ -295,7 +322,7 @@ class Sequence(object):
                 file.write("%s%s is a missing required element of %s\n" % ("    " * indent, element.name, self.__class__.__name__))
                 continue
 
-            if element.klass in _sequence_of_classes:
+            if (element.klass in _sequence_of_classes) or (element.klass in _list_of_classes):
                 file.write("%s%s\n" % ("    " * indent, element.name))
                 helper = element.klass(value)
                 helper.debug_contents(indent+1, file, _ids)
@@ -313,6 +340,7 @@ class Sequence(object):
     def dict_contents(self, use_dict=None, as_class=dict):
         """Return the contents of an object as a dict."""
         if _debug: Sequence._debug("dict_contents use_dict=%r as_class=%r", use_dict, as_class)
+        global _sequence_of_classes, _list_of_classes
 
         # make/extend the dictionary of content
         if use_dict is None:
@@ -324,7 +352,7 @@ class Sequence(object):
             if value is None:
                 continue
 
-            if element.klass in _sequence_of_classes:
+            if (element.klass in _sequence_of_classes) or (element.klass in _list_of_classes):
                 helper = element.klass(value)
                 mapped_value = helper.dict_contents(as_class=as_class)
 
@@ -406,6 +434,9 @@ def SequenceOf(klass):
         def __getitem__(self, item):
             return self.value[item]
 
+        def __iter__(self):
+            return iter(self.value)
+
         def encode(self, taglist):
             if _debug: _SequenceOf._debug("(%r)encode %r", self.__class__.__name__, taglist)
             for value in self.value:
@@ -492,6 +523,158 @@ def SequenceOf(klass):
     # return this new type
     return _SequenceOf
 
+#
+#   List
+#
+
+class List(object):
+    pass
+
+#
+#   ListOf
+#
+
+_list_of_map = {}
+_list_of_classes = {}
+
+@bacpypes_debugging
+def ListOf(klass):
+    """Function to return a class that can encode and decode a list of
+    some other type."""
+    if _debug: ListOf._debug("ListOf %r", klass)
+
+    global _list_of_map
+    global _list_of_classes, _array_of_classes
+
+    # if this has already been built, return the cached one
+    if klass in _list_of_map:
+        if _debug: SequenceOf._debug("    - found in cache")
+        return _list_of_map[klass]
+
+    # no ListOf(ListOf(...)) allowed
+    if klass in _list_of_classes:
+        raise TypeError("nested lists disallowed")
+    # no ListOf(ArrayOf(...)) allowed
+    if klass in _array_of_classes:
+        raise TypeError("lists of arrays disallowed")
+
+    # define a generic class for lists
+    @bacpypes_debugging
+    class _ListOf(List):
+
+        subtype = None
+
+        def __init__(self, value=None):
+            if _debug: _ListOf._debug("(%r)__init__ %r (subtype=%r)", self.__class__.__name__, value, self.subtype)
+
+            if value is None:
+                self.value = []
+            elif isinstance(value, list):
+                self.value = value
+            else:
+                raise TypeError("invalid constructor datatype")
+
+        def append(self, value):
+            if issubclass(self.subtype, Atomic):
+                pass
+            elif issubclass(self.subtype, AnyAtomic) and not isinstance(value, Atomic):
+                raise TypeError("instance of an atomic type required")
+            elif not isinstance(value, self.subtype):
+                raise TypeError("%s value required" % (self.subtype.__name__,))
+            self.value.append(value)
+
+        def __len__(self):
+            return len(self.value)
+
+        def __getitem__(self, item):
+            return self.value[item]
+
+        def encode(self, taglist):
+            if _debug: _ListOf._debug("(%r)encode %r", self.__class__.__name__, taglist)
+            for value in self.value:
+                if issubclass(self.subtype, (Atomic, AnyAtomic)):
+                    # a helper cooperates between the atomic value and the tag
+                    helper = self.subtype(value)
+
+                    # build a tag and encode the data into it
+                    tag = Tag()
+                    helper.encode(tag)
+
+                    # now encode the tag
+                    taglist.append(tag)
+                elif isinstance(value, self.subtype):
+                    # it must have its own encoder
+                    value.encode(taglist)
+                else:
+                    raise TypeError("%s must be a %s" % (value, self.subtype.__name__))
+
+        def decode(self, taglist):
+            if _debug: _ListOf._debug("(%r)decode %r", self.__class__.__name__, taglist)
+
+            while len(taglist) != 0:
+                tag = taglist.Peek()
+                if tag.tagClass == Tag.closingTagClass:
+                    return
+
+                if issubclass(self.subtype, (Atomic, AnyAtomic)):
+                    if _debug: _ListOf._debug("    - building helper: %r %r", self.subtype, tag)
+                    taglist.Pop()
+
+                    # a helper cooperates between the atomic value and the tag
+                    helper = self.subtype(tag)
+
+                    # save the value
+                    self.value.append(helper.value)
+                else:
+                    if _debug: _ListOf._debug("    - building value: %r", self.subtype)
+                    # build an element
+                    value = self.subtype()
+
+                    # let it decode itself
+                    value.decode(taglist)
+
+                    # save what was built
+                    self.value.append(value)
+
+        def debug_contents(self, indent=1, file=sys.stdout, _ids=None):
+            i = 0
+            for value in self.value:
+                if issubclass(self.subtype, (Atomic, AnyAtomic)):
+                    file.write("%s[%d] = %r\n" % ("    " * indent, i, value))
+                elif isinstance(value, self.subtype):
+                    file.write("%s[%d]" % ("    " * indent, i))
+                    value.debug_contents(indent+1, file, _ids)
+                else:
+                    file.write("%s[%d] %s must be a %s" % ("    " * indent, i, value, self.subtype.__name__))
+                i += 1
+
+        def dict_contents(self, use_dict=None, as_class=dict):
+            # return sequences as arrays
+            mapped_value = []
+
+            for value in self.value:
+                if issubclass(self.subtype, Atomic):
+                    mapped_value.append(value)              ### ambiguous
+                elif issubclass(self.subtype, AnyAtomic):
+                    mapped_value.append(value.value)        ### ambiguous
+                elif isinstance(value, self.subtype):
+                    mapped_value.append(value.dict_contents(as_class=as_class))
+
+            # return what we built
+            return mapped_value
+
+    # constrain it to a list of a specific type of item
+    setattr(_ListOf, 'subtype', klass)
+    _ListOf.__name__ = 'ListOf' + klass.__name__
+    if _debug: ListOf._debug("    - build this class: %r", _ListOf)
+
+    # cache this type
+    _list_of_map[klass] = _ListOf
+    _list_of_classes[_ListOf] = 1
+
+    # return this new type
+    return _ListOf
+
 #
 #   Array
 #
@@ -509,15 +692,33 @@ class Array(object):
 _array_of_map = {}
 _array_of_classes = {}
 
-def ArrayOf(klass):
+def ArrayOf(klass, fixed_length=None, prototype=None):
     """Function to return a class that can encode and decode a list of
     some other type."""
     global _array_of_map
     global _array_of_classes, _sequence_of_classes
 
+    # check the parameters for consistency
+    if issubclass(klass, Atomic):
+        if prototype is None:
+            pass
+        elif not klass.is_valid(prototype):
+            raise ValueError("prototype %r not valid for %s" % (prototype, klass.__name__))
+    else:
+        if prototype is None:
+            ### TODO This should be an error, a prototype should always be
+            ### required for non-atomic types, even if it's only klass()
+            ### for a default object which will be deep copied
+            pass
+        elif not isinstance(prototype, klass):
+            raise ValueError("prototype %r not valid for %s" % (prototype, klass.__name__))
+
+    # build a signature of the parameters
+    array_signature = (klass, fixed_length, prototype)
+
     # if this has already been built, return the cached one
-    if klass in _array_of_map:
-        return _array_of_map[klass]
+    if array_signature in _array_of_map:
+        return _array_of_map[array_signature]
 
     # no ArrayOf(ArrayOf(...)) allowed
     if klass in _array_of_classes:
@@ -531,23 +732,60 @@ def ArrayOf(klass):
     class ArrayOf(Array):
 
         subtype = None
+        fixed_length = None
+        prototype = None
 
         def __init__(self, value=None):
             if value is None:
                 self.value = [0]
+                if self.fixed_length is not None:
+                    self.fix_length(self.fixed_length)
+
             elif isinstance(value, list):
+                if (self.fixed_length is not None) and (len(value) != self.fixed_length):
+                    raise ValueError("invalid array length")
+
                 self.value = [len(value)]
                 self.value.extend(value)
             else:
                 raise TypeError("invalid constructor datatype")
 
+        def fix_length(self, new_length):
+            if len(self.value) > new_length + 1:
+                # trim off the excess
+                del self.value[new_length + 1:]
+            elif len(self.value) < new_length + 1:
+                # how many do we need
+                element_count = new_length - len(self.value) + 1
+
+                # extend or append
+                if issubclass(self.subtype, Atomic):
+                    if self.prototype is None:
+                        extend_value = self.subtype().value
+                    else:
+                        extend_value = self.prototype
+                    self.value.extend( [extend_value] * element_count )
+                else:
+                    for i in range(element_count):
+                        if self.prototype is None:
+                            append_value = self.subtype()
+                        else:
+                            append_value = _deepcopy(self.prototype)
+                        self.value.append(append_value)
+
+            self.value[0] = new_length
+
         def append(self, value):
+            if self.fixed_length is not None:
+                raise TypeError("fixed length array")
+
             if issubclass(self.subtype, Atomic):
                 pass
             elif issubclass(self.subtype, AnyAtomic) and not isinstance(value, Atomic):
                 raise TypeError("instance of an atomic type required")
             elif not isinstance(value, self.subtype):
                 raise TypeError("%s value required" % (self.subtype.__name__,))
+
             self.value.append(value)
             self.value[0] = len(self.value) - 1
 
@@ -568,23 +806,19 @@ def ArrayOf(klass):
 
             # special length handling for index 0
             if item == 0:
-                if value < self.value[0]:
-                    # trim
-                    self.value = self.value[0:value + 1]
-                elif value > self.value[0]:
-                    # extend
-                    if issubclass(self.subtype, Atomic):
-                        self.value.extend( [self.subtype().value] * (value - self.value[0]) )
-                    else:
-                        for i in range(value - self.value[0]):
-                            self.value.append(self.subtype())
-                else:
+                if (self.fixed_length is not None):
+                    if (value != self.value[0]):
+                        raise TypeError("fixed length array")
                     return
-                self.value[0] = value
+
+                self.fix_length(value)
             else:
                 self.value[item] = value
 
         def __delitem__(self, item):
+            if self.fixed_length is not None:
+                raise TypeError("fixed length array")
+
             # no wrapping index
             if (item < 1) or (item > self.value[0]):
                 raise IndexError("index out of range")
@@ -593,6 +827,9 @@ def ArrayOf(klass):
             del self.value[item]
             self.value[0] -= 1
 
+        def __iter__(self):
+            return iter(self.value[1:])
+
         def index(self, value):
             # only search through values
             for i in range(1, self.value[0] + 1):
@@ -603,6 +840,9 @@ def ArrayOf(klass):
             raise ValueError("%r not in array" % (value,))
 
         def remove(self, item):
+            if self.fixed_length is not None:
+                raise TypeError("fixed length array")
+
             # find the index of the item and delete it
             indx = self.index(item)
             self.__delitem__(indx)
@@ -631,7 +871,7 @@ def ArrayOf(klass):
             if _debug: ArrayOf._debug("(%r)decode %r", self.__class__.__name__, taglist)
 
             # start with an empty array
-            self.value = [0]
+            new_value = []
 
             while len(taglist) != 0:
                 tag = taglist.Peek()
@@ -646,7 +886,7 @@ def ArrayOf(klass):
                     helper = self.subtype(tag)
 
                     # save the value
-                    self.value.append(helper.value)
+                    new_value.append(helper.value)
                 else:
                     if _debug: ArrayOf._debug("    - building value: %r", self.subtype)
                     # build an element
@@ -656,10 +896,15 @@ def ArrayOf(klass):
                     value.decode(taglist)
 
                     # save what was built
-                    self.value.append(value)
+                    new_value.append(value)
+
+            # check the length
+            if self.fixed_length is not None:
+                if self.fixed_length != len(new_value):
+                    raise ValueError("invalid array length")
 
             # update the length
-            self.value[0] = len(self.value) - 1
+            self.value = [len(new_value)] + new_value
 
         def encode_item(self, item, taglist):
             if _debug: ArrayOf._debug("(%r)encode_item %r %r", self.__class__.__name__, item, taglist)
@@ -756,10 +1001,14 @@ def ArrayOf(klass):
 
     # constrain it to a list of a specific type of item
     setattr(ArrayOf, 'subtype', klass)
+    setattr(ArrayOf, 'fixed_length', fixed_length)
+    setattr(ArrayOf, 'prototype', prototype)
+
+    # update the name
     ArrayOf.__name__ = 'ArrayOf' + klass.__name__
 
     # cache this type
-    _array_of_map[klass] = ArrayOf
+    _array_of_map[array_signature] = ArrayOf
     _array_of_classes[ArrayOf] = 1
 
     # return this new type
@@ -844,6 +1093,7 @@ class Choice(object):
 
     def decode(self, taglist):
         if _debug: Choice._debug("(%r)decode %r", self.__class__.__name__, taglist)
+        global _sequence_of_classes, _list_of_classes
 
         # peek at the element
         tag = taglist.Peek()
@@ -860,7 +1110,7 @@ class Choice(object):
             if _debug: Choice._debug("    - checking choice: %s", element.name)
 
             # check for a sequence element
-            if element.klass in _sequence_of_classes:
+            if (element.klass in _sequence_of_classes) or (element.klass in _list_of_classes):
                 # check for context encoding
                 if element.context is None:
                     raise NotImplementedError("choice of a SequenceOf must be context encoded")
@@ -962,7 +1212,7 @@ class Choice(object):
 
     def dict_contents(self, use_dict=None, as_class=dict):
         """Return the contents of an object as a dict."""
-        if _debug: _log.debug("dict_contents use_dict=%r as_class=%r", use_dict, as_class)
+        if _debug: Choice._debug("dict_contents use_dict=%r as_class=%r", use_dict, as_class)
 
         # make/extend the dictionary of content
         if use_dict is None:
@@ -1045,9 +1295,10 @@ class Any:
     def cast_out(self, klass):
         """Interpret the content as a particular class."""
         if _debug: Any._debug("cast_out %r", klass)
+        global _sequence_of_classes, _list_of_classes
 
         # check for a sequence element
-        if klass in _sequence_of_classes:
+        if (klass in _sequence_of_classes) or (klass in _list_of_classes):
             # build a sequence helper
             helper = klass()
 
@@ -1153,7 +1404,7 @@ class Any:
 #
 
 @bacpypes_debugging
-class AnyAtomic:
+class AnyAtomic(Atomic):
 
     def __init__(self, arg=None):
         if _debug: AnyAtomic._debug("__init__ %r", arg)
@@ -1184,8 +1435,13 @@ class AnyAtomic:
         # get the data
         self.value = tag.app_to_object()
 
+    @classmethod
+    def is_valid(cls, arg):
+        """Return True if arg is valid value for the class."""
+        return isinstance(arg, Atomic) and not isinstance(arg, AnyAtomic)
+
     def __str__(self):
-        return "AnyAtomic(%s)" % (str(self.value), )
+        return "%s(%s)" % (self.__class__.__name__, str(self.value))
 
     def __repr__(self):
         desc = self.__module__ + '.' + self.__class__.__name__
@@ -1196,3 +1452,45 @@ class AnyAtomic:
 
         return '<' + desc + ' instance at 0x%08x' % (id(self),) + '>'
 
+#
+#   SequenceOfAny
+#
+
+@bacpypes_debugging
+class SequenceOfAny(Any):
+
+    def cast_in(self, element):
+        """encode the element into the internal tag list."""
+        if _debug: SequenceOfAny._debug("cast_in %r", element)
+
+        # make sure it is a list
+        if not isinstance(element, List):
+            raise EncodingError("%r is not a list" % (element,))
+
+        t = TagList()
+        element.encode(t)
+
+        self.tagList.extend(t.tagList)
+
+    def cast_out(self, klass):
+        """Interpret the content as a particular class."""
+        if _debug: SequenceOfAny._debug("cast_out %r", klass)
+
+        # make sure it is a list
+        if not issubclass(klass, List):
+            raise DecodingError("%r is not a list" % (klass,))
+
+        # build a helper
+        helper = klass()
+
+        # make a copy of the tag list
+        t = TagList(self.tagList[:])
+
+        # let it decode itself
+        helper.decode(t)
+
+        # make sure everything was consumed
+        if len(t) != 0:
+            raise DecodingError("incomplete cast")
+
+        return helper.value
