@@ -14,6 +14,9 @@ from .debugging import ModuleLogger, btox
 from .errors import DecodingError, InvalidTag, InvalidParameterDatatype
 from .pdu import PDUData
 
+# import the task manager to get the "current" date and time
+from .task import TaskManager as _TaskManager
+
 # some debugging
 _debug = 0
 _log = ModuleLogger(globals())
@@ -453,6 +456,8 @@ class Atomic(object):
     _app_tag = None
 
     def __cmp__(self, other):
+        # sys.stderr.write("__cmp__ %r %r\n" % (self, other))
+
         # hoop jump it
         if not isinstance(other, self.__class__):
             other = self.__class__(other)
@@ -465,6 +470,26 @@ class Atomic(object):
         else:
             return 0
 
+    def __lt__(self, other):
+        # sys.stderr.write("__lt__ %r %r\n" % (self, other))
+
+        # hoop jump it
+        if not isinstance(other, self.__class__):
+            other = self.__class__(other)
+
+        # now compare the values
+        return (self.value < other.value)
+
+    def __eq__(self, other):
+        # sys.stderr.write("__eq__ %r %r\n" % (self, other))
+
+        # hoop jump it
+        if not isinstance(other, self.__class__):
+            other = self.__class__(other)
+
+        # now compare the values
+        return self.value == other.value
+
     @classmethod
     def coerce(cls, arg):
         """Given an arg, return the appropriate value given the class."""
@@ -573,8 +598,10 @@ class Boolean(Atomic):
 class Unsigned(Atomic):
 
     _app_tag = Tag.unsignedAppTag
+    _low_limit = 0
+    _high_limit = None
 
-    def __init__(self,arg = None):
+    def __init__(self, arg=None):
         self.value = 0L
 
         if arg is None:
@@ -582,14 +609,16 @@ class Unsigned(Atomic):
         elif isinstance(arg, Tag):
             self.decode(arg)
         elif isinstance(arg, int):
-            if (arg < 0):
-                raise ValueError("unsigned integer required")
+            if not self.is_valid(arg):
+                raise ValueError("value out of range")
             self.value = long(arg)
         elif isinstance(arg, long):
-            if (arg < 0):
-                raise ValueError("unsigned integer required")
+            if not self.is_valid(arg):
+                raise ValueError("value out of range")
             self.value = arg
         elif isinstance(arg, Unsigned):
+            if not self.is_valid(arg.value):
+                raise ValueError("value out of range")
             self.value = arg.value
         else:
             raise TypeError("invalid constructor datatype")
@@ -622,10 +651,24 @@ class Unsigned(Atomic):
     @classmethod
     def is_valid(cls, arg):
         """Return True if arg is valid value for the class."""
-        return isinstance(arg, (int, long)) and (not isinstance(arg, bool)) and (arg >= 0)
+        if not isinstance(arg, (int, long)) or isinstance(arg, bool):
+            return False
+        if (arg < cls._low_limit):
+            return False
+        if (cls._high_limit is not None) and (arg > cls._high_limit):
+            return False
+        return True
 
     def __str__(self):
-        return "Unsigned(%s)" % (self.value, )
+        return "%s(%s)" % (self.__class__.__name__, self.value)
+
+class Unsigned8(Unsigned):
+    _low_limit = 0
+    _high_limit = 255
+
+class Unsigned16(Unsigned):
+    _low_limit = 0
+    _high_limit = 65535
 
 #
 #   Integer
@@ -1344,6 +1387,9 @@ class Date(Atomic):
         elif isinstance(arg, Date):
             self.value = arg.value
 
+        elif isinstance(arg, float):
+            self.now(arg)
+
         else:
             raise TypeError("invalid constructor datatype")
 
@@ -1372,11 +1418,31 @@ class Date(Atomic):
         # put it back together
         self.value = (year, month, day, day_of_week)
 
-    def now(self):
-        tup = time.localtime()
+    def now(self, when=None):
+        """Set the current value to the correct tuple based on the seconds
+        since the epoch.  If 'when' is not provided, get the current time
+        from the task manager.
+        """
+        if when is None:
+            when = _TaskManager().get_time()
+        tup = time.localtime(when)
+
         self.value = (tup[0]-1900, tup[1], tup[2], tup[6] + 1)
+
         return self
 
+    def __float__(self):
+        """Convert to seconds since the epoch."""
+        # rip apart the value
+        year, month, day, day_of_week = self.value
+
+        # check for special values
+        if (year == 255) or (month in _special_mon_inv) or (day in _special_day_inv):
+            raise ValueError("no wildcard values")
+
+        # convert to time.time() value
+        return time.mktime( (year + 1900, month, day, 0, 0, 0, 0, 0, -1) )
+
     def encode(self, tag):
         # encode the tag
         tag.set_app_data(Tag.dateAppTag, bytearray(self.value))
@@ -1456,19 +1522,40 @@ class Time(Atomic):
                 tup_list[3] = tup_list[3] * 10
 
             self.value = tuple(tup_list)
+
         elif isinstance(arg, Time):
             self.value = arg.value
+
+        elif isinstance(arg, float):
+            self.now(arg)
+
         else:
             raise TypeError("invalid constructor datatype")
 
-    def now(self):
-        now = time.time()
-        tup = time.localtime(now)
+    def now(self, when=None):
+        """Set the current value to the correct tuple based on the seconds
+        since the epoch.  If 'when' is not provided, get the current time
+        from the task manager.
+        """
+        if when is None:
+            when = _TaskManager().get_time()
+        tup = time.localtime(when)
 
-        self.value = (tup[3], tup[4], tup[5], int((now - int(now)) * 100))
+        self.value = (tup[3], tup[4], tup[5], int((when - int(when)) * 100))
 
         return self
 
+    def __float__(self):
+        """Return the current value as an offset from midnight."""
+        if 255 in self.value:
+            raise ValueError("no wildcard values")
+
+        # rip it apart
+        hour, minute, second, hundredth = self.value
+
+        # put it together
+        return (hour * 3600.0) + (minute * 60.0) + second + (hundredth / 100.0)
+
     def encode(self, tag):
         # encode the tag
         tag.set_app_data(Tag.timeAppTag, bytearray(self.value))
@@ -1524,6 +1611,7 @@ class ObjectType(Enumerated):
         , 'accessUser':35
         , 'accessZone':36
         , 'accumulator':23
+        , 'alertEnrollment':52
         , 'analogInput':0
         , 'analogOutput':1
         , 'analogValue':2
@@ -1533,6 +1621,7 @@ class ObjectType(Enumerated):
         , 'binaryValue':5
         , 'bitstringValue':39
         , 'calendar':6
+        , 'channel':53
         , 'characterstringValue':40
         , 'command':7
         , 'credentialDataInput':37
@@ -1550,6 +1639,7 @@ class ObjectType(Enumerated):
         , 'largeAnalogValue':46
         , 'lifeSafetyPoint':21
         , 'lifeSafetyZone':22
+        , 'lightingOutput':54
         , 'loadControl':28
         , 'loop':12
         , 'multiStateInput':13
@@ -1557,6 +1647,7 @@ class ObjectType(Enumerated):
         , 'multiStateValue':19
         , 'networkSecurity':38
         , 'notificationClass':15
+        , 'notificationForwarder':51
         , 'octetstringValue':47
         , 'positiveIntegerValue':48
         , 'program':16
@@ -1595,6 +1686,16 @@ class ObjectIdentifier(Atomic):
                 self.set_long(long(arg))
             elif isinstance(arg, long):
                 self.set_long(arg)
+            elif isinstance(arg, basestring):
+                try:
+                    objType, objInstance = arg.split(':')
+                    if objType.isdigit():
+                        objType = int(objType)
+                    objInstance = int(objInstance)
+                except:
+                    raise ValueError("invalid format")
+
+                self.set_tuple(objType, objInstance)
             elif isinstance(arg, tuple):
                 self.set_tuple(*arg)
             elif isinstance(arg, ObjectIdentifier):
