@@ -7,12 +7,13 @@ Change Of Value Service
 from ..debugging import bacpypes_debugging, DebugContents, ModuleLogger
 from ..capability import Capability
 
-from ..task import OneShotTask, TaskManager
+from ..core import deferred
+from ..task import OneShotTask, RecurringFunctionTask, TaskManager
 from ..iocb import IOCB
 
 from ..basetypes import DeviceAddress, COVSubscription, PropertyValue, \
     Recipient, RecipientProcess, ObjectPropertyReference
-from ..constructeddata import SequenceOf, Any
+from ..constructeddata import ListOf, Any
 from ..apdu import ConfirmedCOVNotificationRequest, \
     UnconfirmedCOVNotificationRequest, \
     SimpleAckPDU, Error, RejectPDU, AbortPDU
@@ -166,14 +167,31 @@ class COVDetection(DetectionAlgorithm):
         # list of all active subscriptions
         self.cov_subscriptions = SubscriptionList()
 
+    def add_subscription(self, cov):
+        if _debug: COVDetection._debug("add_subscription %r", cov)
+
+        # add it to the subscription list for its object
+        self.cov_subscriptions.append(cov)
+
+    def cancel_subscription(self, cov):
+        if _debug: COVDetection._debug("cancel_subscription %r", cov)
+
+        # cancel the subscription timeout
+        if cov.isScheduled:
+            cov.suspend_task()
+            if _debug: COVDetection._debug("    - task suspended")
+
+        # remove it from the subscription list for its object
+        self.cov_subscriptions.remove(cov)
+
     def execute(self):
         if _debug: COVDetection._debug("execute")
 
         # something changed, send out the notifications
         self.send_cov_notifications()
 
-    def send_cov_notifications(self):
-        if _debug: COVDetection._debug("send_cov_notifications")
+    def send_cov_notifications(self, subscription=None):
+        if _debug: COVDetection._debug("send_cov_notifications %r", subscription)
 
         # check for subscriptions
         if not len(self.cov_subscriptions):
@@ -206,8 +224,15 @@ class COVDetection(DetectionAlgorithm):
             list_of_values.append(property_value)
         if _debug: COVDetection._debug("    - list_of_values: %r", list_of_values)
 
+        # if the specific subscription was provided, that is the notification
+        # list, otherwise send it to all of them
+        if subscription is not None:
+            notification_list = [subscription]
+        else:
+            notification_list = self.cov_subscriptions
+
         # loop through the subscriptions and send out notifications
-        for cov in self.cov_subscriptions:
+        for cov in notification_list:
             if _debug: COVDetection._debug("    - cov: %s", repr(cov))
 
             # calculate time remaining
@@ -288,20 +313,20 @@ class COVIncrementCriteria(COVDetection):
             self.previous_reported_value = old_value
 
         # see if it changed enough to trigger reporting
-        value_changed = (new_value <= (self.previous_reported_value - self.covIncrement)) \
-            or (new_value >= (self.previous_reported_value + self.covIncrement))
+        value_changed = (new_value <= (self.previous_reported_value - self.obj.covIncrement)) \
+            or (new_value >= (self.previous_reported_value + self.obj.covIncrement))
         if _debug: COVIncrementCriteria._debug("    - value significantly changed: %r", value_changed)
 
         return value_changed
 
-    def send_cov_notifications(self):
-        if _debug: COVIncrementCriteria._debug("send_cov_notifications")
+    def send_cov_notifications(self, subscription=None):
+        if _debug: COVIncrementCriteria._debug("send_cov_notifications %r", subscription)
 
         # when sending out notifications, keep the current value
         self.previous_reported_value = self.presentValue
 
         # continue
-        COVDetection.send_cov_notifications(self)
+        COVDetection.send_cov_notifications(self, subscription)
 
 
 class AccessDoorCriteria(COVDetection):
@@ -368,17 +393,81 @@ class LoadControlCriteria(COVDetection):
         )
 
 
-class PulseConverterCriteria(COVDetection):
+@bacpypes_debugging
+class PulseConverterCriteria(COVIncrementCriteria):
 
     properties_tracked = (
         'presentValue',
         'statusFlags',
+        'covPeriod',
         )
     properties_reported = (
         'presentValue',
         'statusFlags',
         )
 
+    def __init__(self, obj):
+        if _debug: PulseConverterCriteria._debug("__init__ %r", obj)
+        COVIncrementCriteria.__init__(self, obj)
+
+        # check for a period
+        if self.covPeriod == 0:
+            if _debug: PulseConverterCriteria._debug("    - no periodic notifications")
+            self.cov_period_task = None
+        else:
+            if _debug: PulseConverterCriteria._debug("    - covPeriod: %r", self.covPeriod)
+            self.cov_period_task = RecurringFunctionTask(self.covPeriod * 1000, self.send_cov_notifications)
+            if _debug: PulseConverterCriteria._debug("    - cov period task created")
+
+    def add_subscription(self, cov):
+        if _debug: PulseConverterCriteria._debug("add_subscription %r", cov)
+
+        # let the parent classes do their thing
+        COVIncrementCriteria.add_subscription(self, cov)
+
+        # if there is a COV period task, install it
+        if self.cov_period_task:
+            self.cov_period_task.install_task()
+            if _debug: PulseConverterCriteria._debug("    - cov period task installed")
+
+    def cancel_subscription(self, cov):
+        if _debug: PulseConverterCriteria._debug("cancel_subscription %r", cov)
+
+        # let the parent classes do their thing
+        COVIncrementCriteria.cancel_subscription(self, cov)
+
+        # if there are no more subscriptions, cancel the task
+        if not len(self.cov_subscriptions):
+            if self.cov_period_task and self.cov_period_task.isScheduled:
+                self.cov_period_task.suspend_task()
+                if _debug: PulseConverterCriteria._debug("    - cov period task suspended")
+                self.cov_period_task = None
+
+    @monitor_filter('covPeriod')
+    def cov_period_filter(self, old_value, new_value):
+        if _debug: PulseConverterCriteria._debug("cov_period_filter %r %r", old_value, new_value)
+
+        # check for an old period
+        if old_value != 0:
+            if self.cov_period_task.isScheduled:
+                self.cov_period_task.suspend_task()
+                if _debug: PulseConverterCriteria._debug("    - canceled old task")
+            self.cov_period_task = None
+
+        # check for a new period
+        if new_value != 0:
+            self.cov_period_task = RecurringFunctionTask(new_value * 1000, self.cov_period_x)
+            self.cov_period_task.install_task()
+            if _debug: PulseConverterCriteria._debug("    - new task created and installed")
+
+        return False
+
+    def send_cov_notifications(self, subscription=None):
+        if _debug: PulseConverterCriteria._debug("send_cov_notifications %r", subscription)
+
+        # pass along to the parent class as if something changed
+        COVIncrementCriteria.send_cov_notifications(self, subscription)
+
 
 # mapping from object type to appropriate criteria class
 criteria_type_map = {
@@ -420,7 +509,7 @@ class ActiveCOVSubscriptions(Property):
 
     def __init__(self):
         Property.__init__(
-            self, 'activeCovSubscriptions', SequenceOf(COVSubscription),
+            self, 'activeCovSubscriptions', ListOf(COVSubscription),
             default=None, optional=True, mutable=False,
             )
 
@@ -432,51 +521,50 @@ class ActiveCOVSubscriptions(Property):
         if _debug: ActiveCOVSubscriptions._debug("    - current_time: %r", current_time)
 
         # start with an empty sequence
-        cov_subscriptions = SequenceOf(COVSubscription)()
+        cov_subscriptions = ListOf(COVSubscription)()
 
-        # loop through the object and detection list
-        for obj, cov_detection in obj._app.cov_detections.items():
-            for cov in cov_detection.cov_subscriptions:
-                # calculate time remaining
-                if not cov.lifetime:
-                    time_remaining = 0
-                else:
-                    time_remaining = int(cov.taskTime - current_time)
+        # loop through the subscriptions
+        for cov in obj._app.subscriptions():
+            # calculate time remaining
+            if not cov.lifetime:
+                time_remaining = 0
+            else:
+                time_remaining = int(cov.taskTime - current_time)
 
-                    # make sure it is at least one second
-                    if not time_remaining:
-                        time_remaining = 1
+                # make sure it is at least one second
+                if not time_remaining:
+                    time_remaining = 1
 
-                recipient = Recipient(
-                    address=DeviceAddress(
-                        networkNumber=cov.client_addr.addrNet or 0,
-                        macAddress=cov.client_addr.addrAddr,
-                        ),
-                    )
-                if _debug: ActiveCOVSubscriptions._debug("    - recipient: %r", recipient)
-                if _debug: ActiveCOVSubscriptions._debug("    - client MAC address: %r", cov.client_addr.addrAddr)
+            recipient = Recipient(
+                address=DeviceAddress(
+                    networkNumber=cov.client_addr.addrNet or 0,
+                    macAddress=cov.client_addr.addrAddr,
+                    ),
+                )
+            if _debug: ActiveCOVSubscriptions._debug("    - recipient: %r", recipient)
+            if _debug: ActiveCOVSubscriptions._debug("    - client MAC address: %r", cov.client_addr.addrAddr)
 
-                recipient_process = RecipientProcess(
-                    recipient=recipient,
-                    processIdentifier=cov.proc_id,
-                    )
-                if _debug: ActiveCOVSubscriptions._debug("    - recipient_process: %r", recipient_process)
+            recipient_process = RecipientProcess(
+                recipient=recipient,
+                processIdentifier=cov.proc_id,
+                )
+            if _debug: ActiveCOVSubscriptions._debug("    - recipient_process: %r", recipient_process)
 
-                cov_subscription = COVSubscription(
-                    recipient=recipient_process,
-                    monitoredPropertyReference=ObjectPropertyReference(
-                        objectIdentifier=cov.obj_id,
-                        propertyIdentifier=cov_detection.monitored_property_reference,
-                        ),
-                    issueConfirmedNotifications=cov.confirmed,
-                    timeRemaining=time_remaining,
-                    )
-                if hasattr(cov_detection, 'covIncrement'):
-                    cov_subscription.covIncrement = cov_detection.covIncrement
-                if _debug: ActiveCOVSubscriptions._debug("    - cov_subscription: %r", cov_subscription)
+            cov_subscription = COVSubscription(
+                recipient=recipient_process,
+                monitoredPropertyReference=ObjectPropertyReference(
+                    objectIdentifier=cov.obj_id,
+                    propertyIdentifier=cov_detection.monitored_property_reference,
+                    ),
+                issueConfirmedNotifications=cov.confirmed,
+                timeRemaining=time_remaining,
+                )
+            if hasattr(cov_detection, 'covIncrement'):
+                cov_subscription.covIncrement = cov_detection.covIncrement
+            if _debug: ActiveCOVSubscriptions._debug("    - cov_subscription: %r", cov_subscription)
 
-                # add the list
-                cov_subscriptions.append(cov_subscription)
+            # add the list
+            cov_subscriptions.append(cov_subscription)
 
         return cov_subscriptions
 
@@ -506,22 +594,17 @@ class ChangeOfValueServices(Capability):
     def add_subscription(self, cov):
         if _debug: ChangeOfValueServices._debug("add_subscription %r", cov)
 
-        # add it to the subscription list for its object
-        self.cov_detections[cov.obj_ref].cov_subscriptions.append(cov)
+        # let the detection algorithm know this is a new or additional subscription
+        self.cov_detections[cov.obj_ref].add_subscription(cov)
 
     def cancel_subscription(self, cov):
         if _debug: ChangeOfValueServices._debug("cancel_subscription %r", cov)
 
-        # cancel the subscription timeout
-        if cov.isScheduled:
-            cov.suspend_task()
-            if _debug: ChangeOfValueServices._debug("    - task suspended")
-
         # get the detection algorithm object
         cov_detection = self.cov_detections[cov.obj_ref]
 
-        # remove it from the subscription list for its object
-        cov_detection.cov_subscriptions.remove(cov)
+        # let the detection algorithm know this subscription is going away
+        cov_detection.cancel_subscription(cov)
 
         # if the detection algorithm doesn't have any subscriptions, remove it
         if not len(cov_detection.cov_subscriptions):
@@ -533,6 +616,15 @@ class ChangeOfValueServices(Capability):
             # delete it from the object map
             del self.cov_detections[cov.obj_ref]
 
+    def subscriptions(self):
+        """Generator for the active subscriptions."""
+        if _debug: ChangeOfValueServices._debug("subscriptions")
+
+        # loop through the object and detection list
+        for obj, cov_detection in self.cov_detections.items():
+            for cov in cov_detection.cov_subscriptions:
+                yield cov
+
     def cov_notification(self, cov, request):
         if _debug: ChangeOfValueServices._debug("cov_notification %s %s", str(cov), str(request))
 
@@ -647,3 +739,10 @@ class ChangeOfValueServices(Capability):
 
         # return the result
         self.response(response)
+
+        # if the subscription is not being canceled, it is new or renewed,
+        # so send it a notification when you get a chance.
+        if not cancel_subscription:
+            if _debug: ChangeOfValueServices._debug("    - send a notification")
+            deferred(cov_detection.send_cov_notifications, cov)
+
